@@ -48,10 +48,10 @@
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
-#include <sys/stat.h>
 
 #include <jimautoconf.h>
 #include <jim-subcmd.h>
+#include <jimiocompat.h>
 
 #ifdef HAVE_UTIMES
 #include <sys/time.h>
@@ -68,15 +68,40 @@
 #endif
 
 # ifndef MAXPATHLEN
+# ifdef PATH_MAX
+# define MAXPATHLEN PATH_MAX
+# else
 # define MAXPATHLEN JIM_PATH_LEN
 # endif
+# endif
 
 #if defined(__MINGW32__) || defined(__MSYS__) || defined(_MSC_VER)
 #define ISWINDOWS 1
+/* Even if we have symlink it isn't compatible enought to use */
+#undef HAVE_SYMLINK
 #else
 #define ISWINDOWS 0
 #endif
 
+/* extract nanosecond resolution mtime from struct stat */
+#if defined(HAVE_STRUCT_STAT_ST_MTIMESPEC)
+    #define STAT_MTIME_US(STAT) ((STAT).st_mtimespec.tv_sec * 1000000ll + (STAT).st_mtimespec.tv_nsec / 1000)
+#elif defined(HAVE_STRUCT_STAT_ST_MTIM)
+    #define STAT_MTIME_US(STAT) ((STAT).st_mtim.tv_sec * 1000000ll + (STAT).st_mtim.tv_nsec / 1000)
+#endif
+
+/* On windows, convert backslashes to forward slashes (in place) in the null terminated path */
+static void JimFixPath(char *path)
+{
+    if (ISWINDOWS) {
+        /* Try to keep backslashes out of paths */
+        char *p = path;
+        while ((p = strchr(p, '\\')) != NULL) {
+            *p++ = '/';
+        }
+    }
+}
+
 /*
  *----------------------------------------------------------------------
  *
@@ -133,11 +158,13 @@ static const char *JimGetFileType(int mode)
 /*
  *----------------------------------------------------------------------
  *
- * StoreStatData --
+ * Jim_FileStoreStatData --
  *
  *  This is a utility procedure that breaks out the fields of a
  *  "stat" structure and stores them in textual form into the
- *  elements of an associative array.
+ *  elements of an associative array (dict).
+ *  The result is also returned as the Tcl result.
+ *  If varName is NULL, the result is only returned, not stored.
  *
  * Results:
  *  Returns a standard Tcl return value.  If an error occurs then
@@ -154,7 +181,7 @@ static void AppendStatElement(Jim_Interp *interp, Jim_Obj *listObj, const char *
     Jim_ListAppendElement(interp, listObj, Jim_NewIntObj(interp, value));
 }
 
-static int StoreStatData(Jim_Interp *interp, Jim_Obj *varName, const struct stat *sb)
+int Jim_FileStoreStatData(Jim_Interp *interp, Jim_Obj *varName, const jim_stat_t *sb)
 {
     /* Just use a list to store the data */
     Jim_Obj *listObj = Jim_NewListObj(interp, NULL, 0);
@@ -169,6 +196,9 @@ static int StoreStatData(Jim_Interp *interp, Jim_Obj *varName, const struct stat
     AppendStatElement(interp, listObj, "atime", sb->st_atime);
     AppendStatElement(interp, listObj, "mtime", sb->st_mtime);
     AppendStatElement(interp, listObj, "ctime", sb->st_ctime);
+#ifdef STAT_MTIME_US
+    AppendStatElement(interp, listObj, "mtimeus", STAT_MTIME_US(*sb));
+#endif
     Jim_ListAppendElement(interp, listObj, Jim_NewStringObj(interp, "type", -1));
     Jim_ListAppendElement(interp, listObj, Jim_NewStringObj(interp, JimGetFileType((int)sb->st_mode), -1));
 
@@ -205,16 +235,52 @@ static int StoreStatData(Jim_Interp *interp, Jim_Obj *varName, const struct stat
     return JIM_OK;
 }
 
+/**
+ * Give a path of length 'len', returns the length of the path
+ * with any trailing slashes removed.
+ */
+static int JimPathLenNoTrailingSlashes(const char *path, int len)
+{
+    int i;
+    for (i = len; i > 1 && path[i - 1] == '/'; i--) {
+        /* Trailing slash, so remove it */
+        if (ISWINDOWS && path[i - 2] == ':') {
+            /* But on windows, we won't remove the trailing slash from c:/ */
+            break;
+        }
+    }
+    return i;
+}
+
+/**
+ * Give a path in objPtr, returns a new path with any trailing slash removed.
+ * Use Jim_DecrRefCount() on the returned object (which may be identical to objPtr).
+ */
+static Jim_Obj *JimStripTrailingSlashes(Jim_Interp *interp, Jim_Obj *objPtr)
+{
+    int len = Jim_Length(objPtr);
+    const char *path = Jim_String(objPtr);
+    int i = JimPathLenNoTrailingSlashes(path, len);
+    if (i != len) {
+        objPtr = Jim_NewStringObj(interp, path, i);
+    }
+    Jim_IncrRefCount(objPtr);
+    return objPtr;
+}
+
 static int file_cmd_dirname(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    const char *path = Jim_String(argv[0]);
+    Jim_Obj *objPtr = JimStripTrailingSlashes(interp, argv[0]);
+    const char *path = Jim_String(objPtr);
     const char *p = strrchr(path, '/');
 
-    if (!p && path[0] == '.' && path[1] == '.' && path[2] == '\0') {
-        Jim_SetResultString(interp, "..", -1);
-    } else if (!p) {
+    if (!p) {
         Jim_SetResultString(interp, ".", -1);
     }
+    else if (p[1] == 0) {
+        /* Trailing slash so do nothing */
+        Jim_SetResult(interp, objPtr);
+    }
     else if (p == path) {
         Jim_SetResultString(interp, "/", -1);
     }
@@ -223,8 +289,40 @@ static int file_cmd_dirname(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
         Jim_SetResultString(interp, path, p - path + 1);
     }
     else {
-        Jim_SetResultString(interp, path, p - path);
+        /* Strip any trailing slashes from the result */
+        int len = JimPathLenNoTrailingSlashes(path, p - path);
+        Jim_SetResultString(interp, path, len);
     }
+    Jim_DecrRefCount(interp, objPtr);
+    return JIM_OK;
+}
+
+static int file_cmd_split(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
+{
+    Jim_Obj *listObj = Jim_NewListObj(interp, NULL, 0);
+    const char *path = Jim_String(argv[0]);
+
+    if (*path == '/') {
+        Jim_ListAppendElement(interp, listObj, Jim_NewStringObj(interp, "/", 1));
+    }
+
+    while (1) {
+        /* Remove leading slashes */
+        while (*path == '/') {
+            path++;
+        }
+        if (*path) {
+            const char *pt = strchr(path, '/');
+            if (pt) {
+                Jim_ListAppendElement(interp, listObj, Jim_NewStringObj(interp, path, pt - path));
+                path = pt;
+                continue;
+            }
+            Jim_ListAppendElement(interp, listObj, Jim_NewStringObj(interp, path, -1));
+        }
+        break;
+    }
+    Jim_SetResult(interp, listObj);
     return JIM_OK;
 }
 
@@ -245,7 +343,8 @@ static int file_cmd_rootname(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
 static int file_cmd_extension(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    const char *path = Jim_String(argv[0]);
+    Jim_Obj *objPtr = JimStripTrailingSlashes(interp, argv[0]);
+    const char *path = Jim_String(objPtr);
     const char *lastSlash = strrchr(path, '/');
     const char *p = strrchr(path, '.');
 
@@ -253,42 +352,54 @@ static int file_cmd_extension(Jim_Interp *interp, int argc, Jim_Obj *const *argv
         p = "";
     }
     Jim_SetResultString(interp, p, -1);
+    Jim_DecrRefCount(interp, objPtr);
     return JIM_OK;
 }
 
 static int file_cmd_tail(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    const char *path = Jim_String(argv[0]);
+    Jim_Obj *objPtr = JimStripTrailingSlashes(interp, argv[0]);
+    const char *path = Jim_String(objPtr);
     const char *lastSlash = strrchr(path, '/');
 
     if (lastSlash) {
         Jim_SetResultString(interp, lastSlash + 1, -1);
     }
     else {
-        Jim_SetResult(interp, argv[0]);
+        Jim_SetResult(interp, objPtr);
     }
+    Jim_DecrRefCount(interp, objPtr);
     return JIM_OK;
 }
 
+#ifndef HAVE_RESTRICT
+#define restrict
+#endif
+
+static char *JimRealPath(const char *restrict path, char *restrict resolved_path, size_t len)
+{
+#if defined(HAVE__FULLPATH)
+    return _fullpath(resolved_path, path, len);
+#elif defined(HAVE_REALPATH)
+    return realpath(path, resolved_path);
+#else
+    return NULL;
+#endif
+}
+
 static int file_cmd_normalize(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-#ifdef HAVE_REALPATH
     const char *path = Jim_String(argv[0]);
-    char *newname = Jim_Alloc(MAXPATHLEN + 1);
+    char *newname = Jim_Alloc(MAXPATHLEN);
 
-    if (realpath(path, newname)) {
+    if (JimRealPath(path, newname, MAXPATHLEN)) {
+        JimFixPath(newname);
         Jim_SetResult(interp, Jim_NewStringObjNoAlloc(interp, newname, -1));
         return JIM_OK;
     }
-    else {
-        Jim_Free(newname);
-        Jim_SetResultFormatted(interp, "can't normalize \"%#s\": %s", argv[0], strerror(errno));
-        return JIM_ERR;
-    }
-#else
-    Jim_SetResultString(interp, "Not implemented", -1);
+    Jim_Free(newname);
+    Jim_SetResultFormatted(interp, "can't normalize \"%#s\": %s", argv[0], strerror(errno));
     return JIM_ERR;
-#endif
 }
 
 static int file_cmd_join(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
@@ -394,8 +505,8 @@ static int file_cmd_delete(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
     int force = Jim_CompareStringImmediate(interp, argv[0], "-force");
 
     if (force || Jim_CompareStringImmediate(interp, argv[0], "--")) {
-        argc++;
-        argv--;
+        argc--;
+        argv++;
     }
 
     while (argc--) {
@@ -459,9 +570,9 @@ static int mkdir_all(char *path)
         }
         /* Maybe it already exists as a directory */
         if (errno == EEXIST) {
-            struct stat sb;
+            jim_stat_t sb;
 
-            if (stat(path, &sb) == 0 && S_ISDIR(sb.st_mode)) {
+            if (Jim_Stat(path, &sb) == 0 && S_ISDIR(sb.st_mode)) {
                 return 0;
             }
             /* Restore errno */
@@ -525,7 +636,12 @@ static int file_cmd_rename(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
             argv[1]);
         return JIM_ERR;
     }
-
+#if ISWINDOWS
+    if (access(dest, F_OK) == 0) {
+        /* Windows won't rename over an existing file */
+        remove(dest);
+    }
+#endif
     if (rename(source, dest) != 0) {
         Jim_SetResultFormatted(interp, "error renaming \"%#s\" to \"%#s\": %s", argv[0], argv[1],
             strerror(errno));
@@ -573,11 +689,11 @@ static int file_cmd_link(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 }
 #endif
 
-static int file_stat(Jim_Interp *interp, Jim_Obj *filename, struct stat *sb)
+static int file_stat(Jim_Interp *interp, Jim_Obj *filename, jim_stat_t *sb)
 {
     const char *path = Jim_String(filename);
 
-    if (stat(path, sb) == -1) {
+    if (Jim_Stat(path, sb) == -1) {
         Jim_SetResultFormatted(interp, "could not read \"%#s\": %s", filename, strerror(errno));
         return JIM_ERR;
     }
@@ -585,7 +701,7 @@ static int file_stat(Jim_Interp *interp, Jim_Obj *filename, struct stat *sb)
 }
 
 #ifdef HAVE_LSTAT
-static int file_lstat(Jim_Interp *interp, Jim_Obj *filename, struct stat *sb)
+static int file_lstat(Jim_Interp *interp, Jim_Obj *filename, jim_stat_t *sb)
 {
     const char *path = Jim_String(filename);
 
@@ -601,7 +717,7 @@ static int file_lstat(Jim_Interp *interp, Jim_Obj *filename, struct stat *sb)
 
 static int file_cmd_atime(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
 
     if (file_stat(interp, argv[0], &sb) != JIM_OK) {
         return JIM_ERR;
@@ -610,30 +726,38 @@ static int file_cmd_atime(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
     return JIM_OK;
 }
 
+/**
+ * Set file atime/mtime to the given time in microseconds since the epoch.
+ */
+static int JimSetFileTimes(Jim_Interp *interp, const char *filename, jim_wide us)
+{
+#ifdef HAVE_UTIMES
+    struct timeval times[2];
+
+    times[1].tv_sec = times[0].tv_sec = us / 1000000;
+    times[1].tv_usec = times[0].tv_usec = us % 1000000;
+
+    if (utimes(filename, times) != 0) {
+        Jim_SetResultFormatted(interp, "can't set time on \"%s\": %s", filename, strerror(errno));
+        return JIM_ERR;
+    }
+    return JIM_OK;
+#else
+    Jim_SetResultString(interp, "Not implemented", -1);
+    return JIM_ERR;
+#endif
+}
+
 static int file_cmd_mtime(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
 
     if (argc == 2) {
-#ifdef HAVE_UTIMES
-        jim_wide newtime;
-        struct timeval times[2];
-
-        if (Jim_GetWide(interp, argv[1], &newtime) != JIM_OK) {
+        jim_wide secs;
+        if (Jim_GetWide(interp, argv[1], &secs) != JIM_OK) {
             return JIM_ERR;
         }
-
-        times[1].tv_sec = times[0].tv_sec = newtime;
-        times[1].tv_usec = times[0].tv_usec = 0;
-
-        if (utimes(Jim_String(argv[0]), times) != 0) {
-            Jim_SetResultFormatted(interp, "can't set time on \"%#s\": %s", argv[0], strerror(errno));
-            return JIM_ERR;
-        }
-#else
-        Jim_SetResultString(interp, "Not implemented", -1);
-        return JIM_ERR;
-#endif
+        return JimSetFileTimes(interp, Jim_String(argv[0]), secs * 1000000);
     }
     if (file_stat(interp, argv[0], &sb) != JIM_OK) {
         return JIM_ERR;
@@ -642,6 +766,26 @@ static int file_cmd_mtime(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
     return JIM_OK;
 }
 
+#ifdef STAT_MTIME_US
+static int file_cmd_mtimeus(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
+{
+    jim_stat_t sb;
+
+    if (argc == 2) {
+        jim_wide us;
+        if (Jim_GetWide(interp, argv[1], &us) != JIM_OK) {
+            return JIM_ERR;
+        }
+        return JimSetFileTimes(interp, Jim_String(argv[0]), us);
+    }
+    if (file_stat(interp, argv[0], &sb) != JIM_OK) {
+        return JIM_ERR;
+    }
+    Jim_SetResultInt(interp, STAT_MTIME_US(sb));
+    return JIM_OK;
+}
+#endif
+
 static int file_cmd_copy(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
     return Jim_EvalPrefix(interp, "file copy", argc, argv);
@@ -649,7 +793,7 @@ static int file_cmd_copy(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
 static int file_cmd_size(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
 
     if (file_stat(interp, argv[0], &sb) != JIM_OK) {
         return JIM_ERR;
@@ -660,7 +804,7 @@ static int file_cmd_size(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
 static int file_cmd_isdirectory(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
     int ret = 0;
 
     if (file_stat(interp, argv[0], &sb) == JIM_OK) {
@@ -672,7 +816,7 @@ static int file_cmd_isdirectory(Jim_Interp *interp, int argc, Jim_Obj *const *ar
 
 static int file_cmd_isfile(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
     int ret = 0;
 
     if (file_stat(interp, argv[0], &sb) == JIM_OK) {
@@ -685,7 +829,7 @@ static int file_cmd_isfile(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 #ifdef HAVE_GETEUID
 static int file_cmd_owned(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
     int ret = 0;
 
     if (file_stat(interp, argv[0], &sb) == JIM_OK) {
@@ -706,7 +850,7 @@ static int file_cmd_readlink(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
     if (linkLength == -1) {
         Jim_Free(linkValue);
-        Jim_SetResultFormatted(interp, "couldn't readlink \"%#s\": %s", argv[0], strerror(errno));
+        Jim_SetResultFormatted(interp, "could not read link \"%#s\": %s", argv[0], strerror(errno));
         return JIM_ERR;
     }
     linkValue[linkLength] = 0;
@@ -717,7 +861,7 @@ static int file_cmd_readlink(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
 static int file_cmd_type(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
 
     if (file_lstat(interp, argv[0], &sb) != JIM_OK) {
         return JIM_ERR;
@@ -729,12 +873,12 @@ static int file_cmd_type(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 #ifdef HAVE_LSTAT
 static int file_cmd_lstat(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
 
     if (file_lstat(interp, argv[0], &sb) != JIM_OK) {
         return JIM_ERR;
     }
-    return StoreStatData(interp, argc == 2 ? argv[1] : NULL, &sb);
+    return Jim_FileStoreStatData(interp, argc == 2 ? argv[1] : NULL, &sb);
 }
 #else
 #define file_cmd_lstat file_cmd_stat
@@ -742,12 +886,12 @@ static int file_cmd_lstat(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
 static int file_cmd_stat(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 {
-    struct stat sb;
+    jim_stat_t sb;
 
     if (file_stat(interp, argv[0], &sb) != JIM_OK) {
         return JIM_ERR;
     }
-    return StoreStatData(interp, argc == 2 ? argv[1] : NULL, &sb);
+    return Jim_FileStoreStatData(interp, argc == 2 ? argv[1] : NULL, &sb);
 }
 
 static const jim_subcmd_type file_command_table[] = {
@@ -765,6 +909,15 @@ static const jim_subcmd_type file_command_table[] = {
         2,
         /* Description: Get or set last modification time */
     },
+#ifdef STAT_MTIME_US
+    {   "mtimeus",
+        "name ?time?",
+        file_cmd_mtimeus,
+        1,
+        2,
+        /* Description: Get or set last modification time in microseconds */
+    },
+#endif
     {   "copy",
         "?-force? source dest",
         file_cmd_copy,
@@ -800,6 +953,13 @@ static const jim_subcmd_type file_command_table[] = {
         1,
         /* Description: Last component of the name */
     },
+    {   "split",
+        "name",
+        file_cmd_split,
+        1,
+        1,
+        /* Description: Split path into components as a list */
+    },
     {   "normalize",
         "name",
         file_cmd_normalize,
@@ -972,14 +1132,7 @@ static int Jim_PwdCmd(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
         Jim_Free(cwd);
         return JIM_ERR;
     }
-    else if (ISWINDOWS) {
-        /* Try to keep backslashes out of paths */
-        char *p = cwd;
-        while ((p = strchr(p, '\\')) != NULL) {
-            *p++ = '/';
-        }
-    }
-
+    JimFixPath(cwd);
     Jim_SetResultString(interp, cwd, -1);
 
     Jim_Free(cwd);
@@ -988,9 +1141,7 @@ static int Jim_PwdCmd(Jim_Interp *interp, int argc, Jim_Obj *const *argv)
 
 int Jim_fileInit(Jim_Interp *interp)
 {
-    if (Jim_PackageProvide(interp, "file", "1.0", JIM_ERRMSG))
-        return JIM_ERR;
-
+    Jim_PackageProvideCheck(interp, "file");
     Jim_CreateCommand(interp, "file", Jim_SubCmdProc, (void *)file_command_table, NULL);
     Jim_CreateCommand(interp, "pwd", Jim_PwdCmd, NULL, NULL);
     Jim_CreateCommand(interp, "cd", Jim_CdCmd, NULL, NULL);
